@@ -22,8 +22,9 @@
 namespace GtkSharp.Generation {
 
 	using System;
+	using System.IO;
 
-	public class LPGen : SimpleGen {
+	public class LPGen : SimpleGen, IAccessor {
 		
 		public LPGen (string ctype) : base (ctype, "long") {}
 
@@ -42,6 +43,16 @@ namespace GtkSharp.Generation {
 		{
 			return "(long) " + var;
 		}
+
+		public void WriteAccessors (StreamWriter sw, string indent, string var)
+		{
+			sw.WriteLine (indent + "get {");
+			sw.WriteLine (indent + "\treturn " + FromNative (var) + ";");
+			sw.WriteLine (indent + "}");
+			sw.WriteLine (indent + "set {");
+			sw.WriteLine (indent + "\t" + var + " = " + CallByName ("value") + ";");
+			sw.WriteLine (indent + "}");
+		}
 	}
 }
 
