@@ -98,6 +98,12 @@ namespace GtkSharp.Generation {
 			}
 		}
 
+		public bool IsDestroyNotify {
+			get {
+				return CType == "GDestroyNotify";
+			}
+		}
+
 		public bool IsLength {
 			get {
 				
@@ -155,9 +161,9 @@ namespace GtkSharp.Generation {
 			}
 		}
 
-		public bool NullOk {
+		public bool Owned {
 			get {
-				return elem.HasAttribute ("null_ok");
+				return elem.GetAttribute ("owned") == "true";
 			}
 		}
 
@@ -182,6 +188,18 @@ namespace GtkSharp.Generation {
 			}
 		}
 
+		string scope;
+		public string Scope {
+			get {
+				if (scope == null)
+					scope = elem.GetAttribute ("scope");
+				return scope;
+			}
+			set {
+				scope = value;
+			}
+		}
+
 		public string CallByName (string call_parm_name)
 		{
 			string call_parm;
@@ -190,14 +208,18 @@ namespace GtkSharp.Generation {
 			else
 				call_parm = SymbolTable.Table.CallByName(CType, call_parm_name);
 			
-			if (NullOk && !CSType.EndsWith ("IntPtr") && !(Generatable is StructBase))
-				call_parm = String.Format ("({0} != null) ? {1} : {2}", call_parm_name, call_parm, Generatable is CallbackGen ? "null" : "IntPtr.Zero");
-
 			if (IsArray)
 				call_parm = call_parm.Replace ("ref ", "");
 
 			return call_parm;
+		}
 
+		public string FromNative (string var)
+		{
+			if (Generatable is HandleBase)
+				return ((HandleBase)Generatable).FromNative (var, Owned);
+			else
+				return Generatable.FromNative (var);
 		}
 
 		public string StudlyName {
@@ -238,12 +260,57 @@ namespace GtkSharp.Generation {
 			}
 		}
 
+		public int VisibleCount {
+			get {
+				int visible = 0;
+				foreach (Parameter p in this) {
+					if (!IsHidden (p))
+						visible++;
+				}
+				return visible;
+			}
+		}
+
 		public Parameter this [int idx] {
 			get {
 				return param_list [idx] as Parameter;
 			}
 		}
 
+		public bool IsHidden (Parameter p)
+		{
+			int idx = param_list.IndexOf (p);
+
+			if (idx > 0 && p.IsLength && this [idx - 1].IsString)
+				return true;
+
+			if (p.IsCount && ((idx > 0 && this [idx - 1].IsArray) ||
+					  (idx < Count - 1 && this [idx + 1].IsArray)))
+				return true;
+
+			if (p.CType == "GError**")
+				return true;
+
+			if (HasCB || HideData) {
+				if (p.IsUserData && (idx == Count - 1))
+                                        return true;
+				if (p.IsUserData && idx > 0 &&
+				    this [idx - 1].Generatable is CallbackGen)
+					return true;
+				if (p.IsDestroyNotify && (idx == Count - 1) &&
+				    this [idx - 1].IsUserData)
+					return true;
+			}
+
+			return false;
+		}
+
+		bool has_cb;
+		public bool HasCB {
+			get { return has_cb; }
+			set { has_cb = value; }
+		}
+
 		bool hide_data;
 		public bool HideData {
 			get { return hide_data; }
@@ -273,7 +340,8 @@ namespace GtkSharp.Generation {
 			if (cleared)
 				return false;
 
-			foreach (Parameter p in param_list) {
+			for (int i = 0; i < param_list.Count; i++) {
+				Parameter p = this [i];
 				
 				if (p.IsEllipsis) {
 					Console.Write("Ellipsis parameter ");
@@ -288,6 +356,14 @@ namespace GtkSharp.Generation {
 					Clear ();
 					return false;
 				}
+
+				if (p.Generatable is CallbackGen) {
+					has_cb = true;
+					if (i == Count - 3 &&
+					    this [i + 1].IsUserData &&
+					    this [i + 2].IsDestroyNotify)
+						p.Scope = "notified";
+				}
 			}
 			
 			return true;
@@ -295,14 +371,25 @@ namespace GtkSharp.Generation {
 
 		public bool IsAccessor {
 			get {
-				return Count == 1 && this [0].PassAs == "out";
+				return VisibleCount == 1 && AccessorParam.PassAs == "out";
+			}
+		}
+
+		public Parameter AccessorParam {
+			get {
+				foreach (Parameter p in this) {
+					if (!IsHidden (p))
+						return p;
+				}
+				return null;
 			}
 		}
 
 		public string AccessorReturnType {
 			get {
-				if (Count > 0)
-					return this [0].CSType;
+				Parameter p = AccessorParam;
+				if (p != null)
+					return p.CSType;
 				else
 					return null;
 			}
@@ -310,8 +397,9 @@ namespace GtkSharp.Generation {
 
 		public string AccessorName {
 			get {
-				if (Count > 0)
-					return this [0].Name;
+				Parameter p = AccessorParam;
+				if (p != null)
+					return p.Name;
 				else
 					return null;
 			}
