@@ -32,30 +32,105 @@ namespace GLib {
 		[DllImport("libglib-2.0-0.dll")]
 		static extern void g_free (IntPtr mem);
 
-		public static string PtrToStringGFree (IntPtr ptr) {
-			string ret = Marshal.PtrToStringAnsi (ptr);
+		public static void Free (IntPtr ptr)
+		{
+			g_free (ptr);
+		}
+
+		[DllImport("libglib-2.0-0.dll")]
+		static extern IntPtr g_filename_to_utf8 (IntPtr mem, int len, IntPtr read, out IntPtr written, out IntPtr error);
+
+		public static string FilenamePtrToString (IntPtr ptr) 
+		{
+			if (ptr == IntPtr.Zero) return null;
+			
+			IntPtr dummy, error;
+			IntPtr utf8 = g_filename_to_utf8 (ptr, -1, IntPtr.Zero, out dummy, out error);
+			if (error != IntPtr.Zero)
+				throw new GLib.GException (error);
+			return Utf8PtrToString (utf8);
+		}
+
+		public static string FilenamePtrToStringGFree (IntPtr ptr) 
+		{
+			string ret = FilenamePtrToString (ptr);
 			g_free (ptr);
 			return ret;
 		}
 
-		[DllImport("libglib-2.0-0.dll")]
-		static extern void g_strfreev (IntPtr mem);
+		[DllImport("glibsharpglue-2")]
+		static extern UIntPtr glibsharp_strlen (IntPtr mem);
+
+		public static string Utf8PtrToString (IntPtr ptr) 
+		{
+			if (ptr == IntPtr.Zero)
+				return null;
+
+			int len = (int) (uint)glibsharp_strlen (ptr);
+			byte[] bytes = new byte [len];
+			Marshal.Copy (ptr, bytes, 0, len);
+			return System.Text.Encoding.UTF8.GetString (bytes);
+		}
+
+		public static string[] Utf8PtrToString (IntPtr[] ptrs) {
+			// The last pointer is a null terminator.
+			string[] ret = new string[ptrs.Length - 1];
+			for (int i = 0; i < ret.Length; i++)
+				ret[i] = Utf8PtrToString (ptrs[i]);
+			return ret;
+		}
+
+		public static string PtrToStringGFree (IntPtr ptr) 
+		{
+			string ret = Utf8PtrToString (ptr);
+			g_free (ptr);
+			return ret;
+		}
 
 		public static string[] PtrToStringGFree (IntPtr[] ptrs) {
 			// The last pointer is a null terminator.
 			string[] ret = new string[ptrs.Length - 1];
 			for (int i = 0; i < ret.Length; i++) {
-				ret[i] = Marshal.PtrToStringAnsi (ptrs[i]);
+				ret[i] = Utf8PtrToString (ptrs[i]);
 				g_free (ptrs[i]);
 			}
 			return ret;
 		}
 
 		[DllImport("libglib-2.0-0.dll")]
-		static extern IntPtr g_strdup (string str);
+		static extern IntPtr g_filename_from_utf8 (IntPtr mem, int len, IntPtr read, out IntPtr written, out IntPtr error);
+
+		public static IntPtr StringToFilenamePtr (string str) 
+		{
+			if (str == null)
+				return IntPtr.Zero;
+
+			IntPtr dummy, error;
+			IntPtr utf8 = StringToPtrGStrdup (str);
+			IntPtr result = g_filename_from_utf8 (utf8, -1, IntPtr.Zero, out dummy, out error);
+			g_free (utf8);
+			if (error != IntPtr.Zero)
+				throw new GException (error);
+
+			return result;
+		}
 
 		public static IntPtr StringToPtrGStrdup (string str) {
-			return g_strdup (str);
+			if (str == null)
+				return IntPtr.Zero;
+			byte[] bytes = System.Text.Encoding.UTF8.GetBytes (str);
+			IntPtr result = g_malloc (new UIntPtr ((ulong)bytes.Length + 1));
+			Marshal.Copy (bytes, 0, result, bytes.Length);
+			Marshal.WriteByte (result, bytes.Length, 0);
+			return result;
+		}
+
+		public static string StringFormat (string format, params object[] args) {
+			string ret = String.Format (format, args);
+			if (ret.IndexOf ('%') == -1)
+				return ret;
+			else
+				return ret.Replace ("%", "%%");
 		}
 
 		// Argv marshalling -- unpleasantly complex, but
@@ -109,6 +184,7 @@ namespace GLib {
 			return buf;
 		}
 
+		[Obsolete ("Use GLib.Argv instead to avoid leaks.")]
 		public static IntPtr ArgvToArrayPtr (string[] args)
 		{
 			if (args.Length == 0)
@@ -149,6 +225,7 @@ namespace GLib {
 			return args;
 		}		
 
+		[Obsolete ("Use GLib.Argv instead to avoid leaks.")]
 		public static string[] ArrayPtrToArgv (IntPtr array, int argc)
 		{
 			if (argc == 0)
@@ -173,7 +250,7 @@ namespace GLib {
 			return local_epoch.AddSeconds ((int)time_t + utc_offset);
 		}
 
-		[DllImport("glibsharpglue")]
+		[DllImport("glibsharpglue-2")]
 		static extern IntPtr gtksharp_unichar_to_utf8_string (uint c);
 
 		public static char GUnicharToChar (uint ucs4_char)
@@ -189,7 +266,7 @@ namespace GLib {
 			return ret [0];
 		}
 
-		[DllImport("glibsharpglue")]
+		[DllImport("glibsharpglue-2")]
 		static extern uint glibsharp_utf16_to_unichar (ushort c);
 
 		public static uint CharToGUnichar (char c)
@@ -197,6 +274,20 @@ namespace GLib {
 			return glibsharp_utf16_to_unichar ((ushort) c);
 		}
 
+		public static IntPtr StructureToPtrAlloc (object o)
+		{
+			IntPtr result = Marshal.AllocHGlobal (Marshal.SizeOf (o));
+			Marshal.StructureToPtr (o, result, false);
+			return result;
+		}
+
+		public static Array ListToArray (ListBase list, System.Type type)
+		{
+			Array result = Array.CreateInstance (type, list.Count);
+			if (list.Count > 0)
+				list.CopyTo (result, 0);
+			return result;
+		}
 	}
 }
 
