@@ -3,7 +3,7 @@
 // Authors: Mike Kestner <mkestner@speakeasy.net>
 //
 // Copyright (c) 2001-2003 Mike Kestner
-// Copyright (c) 2004 Novell, Inc.
+// Copyright (c) 2004-2005 Novell, Inc.
 //
 // This program is free software; you can redistribute it and/or
 // modify it under the terms of version 2 of the Lesser GNU General 
@@ -27,6 +27,7 @@ namespace GLib {
 	using System.ComponentModel;
 	using System.Reflection;
 	using System.Runtime.InteropServices;
+	using System.Text;
 
 	public class Object : IWrapper, IDisposable {
 
@@ -34,7 +35,7 @@ namespace GLib {
 		bool disposed = false;
 		Hashtable data;
 		static Hashtable Objects = new Hashtable();
-		static Queue PendingDestroys = new Queue ();
+		static ArrayList PendingDestroys = new ArrayList ();
 		static bool idle_queued;
 
 		~Object ()
@@ -61,24 +62,29 @@ namespace GLib {
 				if (o._obj == IntPtr.Zero)
 					continue;
 
-				g_object_unref (o._obj);
+				try {
+					g_object_unref (o._obj);
+				} catch (Exception e) {
+					Console.WriteLine ("Exception while disposing a " + o + " in Gtk#");
+					throw e;
+				}
+				Objects.Remove (o._obj);
 				o._obj = IntPtr.Zero;
 			}
 			return false;
 		}
 
-		public void Dispose ()
+		public virtual void Dispose ()
 		{
 			if (disposed)
 				return;
 
 			disposed = true;
-			Objects.Remove (_obj);
 			lock (PendingDestroys){
-				PendingDestroys.Enqueue (this);
+				PendingDestroys.Add (this);
 				lock (typeof (Object)){
 					if (!idle_queued){
-						Idle.Add (new IdleHandler (PerformQueuedUnrefs));
+						Timeout.Add (50, new TimeoutHandler (PerformQueuedUnrefs));
 						idle_queued = true;
 					}
 				}
@@ -91,12 +97,24 @@ namespace GLib {
 
 		public static Object GetObject(IntPtr o, bool owned_ref)
 		{
-			Object obj;
+			if (o == IntPtr.Zero)
+				return null;
+
+			Object obj = null;
 			WeakReference weak_ref = Objects[o] as WeakReference;
-			if (weak_ref != null && weak_ref.IsAlive) {
-				obj = weak_ref.Target as GLib.Object;
+
+			if (weak_ref != null && weak_ref.IsAlive)
+				obj = weak_ref.Target as Object;
+
+			if (obj == null)
+				obj = Objects[o] as Object;
+
+			if (obj != null && obj._obj == o) {
+				lock (PendingDestroys)
+					PendingDestroys.Remove (obj);
 				if (owned_ref)
 					g_object_unref (obj._obj);
+				obj.disposed = false;
 				return obj;
 			}
 
@@ -122,10 +140,7 @@ namespace GLib {
 				if (baseinfo == minfo)
 					continue;
 
-				foreach (object attr in baseinfo.GetCustomAttributes (true)) {
-					if (attr.ToString () != "GLib.DefaultSignalHandlerAttribute")
-						continue;
-
+				foreach (object attr in baseinfo.GetCustomAttributes (typeof (DefaultSignalHandlerAttribute), false)) {
 					DefaultSignalHandlerAttribute sigattr = attr as DefaultSignalHandlerAttribute;
 					MethodInfo connector = sigattr.Type.GetMethod (sigattr.ConnectionMethod, BindingFlags.Static | BindingFlags.NonPublic);
 					object[] parms = new object [1];
@@ -137,16 +152,54 @@ namespace GLib {
 					
 		}
 
-		[DllImport("glibsharpglue")]
-		static extern IntPtr gtksharp_register_type (string name, IntPtr parent_type);
+		private static void InvokeClassInitializers (GType gtype, System.Type t)
+		{
+			object[] parms = {gtype, t};
+			BindingFlags flags = BindingFlags.Static | BindingFlags.NonPublic | BindingFlags.FlattenHierarchy;
+			foreach (MethodInfo minfo in t.GetMethods(flags))
+				if (minfo.IsDefined (typeof (ClassInitializerAttribute), true))
+					minfo.Invoke (null, parms);
+		}
+
+		[DllImport("glibsharpglue-2")]
+		static extern IntPtr gtksharp_register_type (IntPtr name, IntPtr parent_type);
+
+
+		static int type_uid;
+		static string BuildEscapedName (System.Type t)
+		{
+			string qn = t.FullName;
+			// Just a random guess
+			StringBuilder sb = new StringBuilder (20 + qn.Length);
+			sb.Append ("__gtksharp_");
+			sb.Append (type_uid++);
+			sb.Append ("_");
+			foreach (char c in qn) {
+				if ((c >= '0' && c <= '9') || (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
+					sb.Append (c);
+				else if (c == '.')
+					sb.Append ('_');
+				else if ((uint) c <= byte.MaxValue) {
+					sb.Append ('+');
+					sb.Append (((byte) c).ToString ("x2"));
+				} else {
+					sb.Append ('-');
+					sb.Append (((uint) c).ToString ("x4"));
+				}
+			}
+			return sb.ToString ();
+		}
 
 		protected static GType RegisterGType (System.Type t)
 		{
 			GType parent_gtype = LookupGType (t.BaseType);
-			string name = t.FullName.Replace(".", "_");
-			GLib.ObjectManager.RegisterType (name, t.FullName, t.Assembly.GetName().Name);
-			GType gtype = new GType (gtksharp_register_type (name, parent_gtype.Val));
+			string name = BuildEscapedName (t);
+			IntPtr native_name = GLib.Marshaller.StringToPtrGStrdup (name);
+			GType gtype = new GType (gtksharp_register_type (native_name, parent_gtype.Val));
+			GLib.Marshaller.Free (native_name);
+			GLib.GType.Register (gtype, t);
 			ConnectDefaultHandlers (gtype, t);
+			InvokeClassInitializers (gtype, t);
 			g_types[t] = gtype;
 			return gtype;
 		}
@@ -159,7 +212,7 @@ namespace GLib {
 			return LookupGType (GetType ());
 		}
 
-		protected static GType LookupGType (System.Type t)
+		protected internal static GType LookupGType (System.Type t)
 		{
 			if (g_types.Contains (t))
 				return (GType) g_types [t];
@@ -179,17 +232,24 @@ namespace GLib {
 		[DllImport("libgobject-2.0-0.dll")]
 		static extern IntPtr g_object_new (IntPtr gtype, IntPtr dummy);
 
+		[Obsolete]
 		protected Object (GType gtype)
 		{
 			Raw = g_object_new (gtype.Val, IntPtr.Zero);
 		}
 
-		[DllImport("glibsharpglue")]
-		static extern IntPtr gtksharp_object_newv (IntPtr gtype, int n_params, string[] names, GLib.Value[] vals);
+		[DllImport("glibsharpglue-2")]
+		static extern IntPtr gtksharp_object_newv (IntPtr gtype, int n_params, IntPtr[] names, GLib.Value[] vals);
 
 		protected virtual void CreateNativeObject (string[] names, GLib.Value[] vals)
 		{
-			Raw = gtksharp_object_newv (LookupGType ().Val, names.Length, names, vals);
+			IntPtr[] native_names = new IntPtr [names.Length];
+			for (int i = 0; i < names.Length; i++)
+				native_names [i] = GLib.Marshaller.StringToPtrGStrdup (names [i]);
+			Raw = gtksharp_object_newv (LookupGType ().Val, names.Length, native_names, vals);
+			Objects [_obj] = this;
+			foreach (IntPtr p in native_names)
+				GLib.Marshaller.Free (p);
 		}
 
 		protected virtual IntPtr Raw {
@@ -206,21 +266,18 @@ namespace GLib {
 			}
 		}	
 
-		[DllImport("glibsharpglue")]
-		private static extern IntPtr gtksharp_get_type_id (IntPtr obj);
-
 		public static GLib.GType GType {
 			get {
 				return GType.Object;
 			}
 		}
 
-		[DllImport("glibsharpglue")]
+		[DllImport("glibsharpglue-2")]
 		static extern IntPtr gtksharp_get_type_name (IntPtr raw);
 
 		protected string TypeName {
 			get {
-				return Marshal.PtrToStringAnsi (gtksharp_get_type_name (Raw));
+				return Marshaller.Utf8PtrToString (gtksharp_get_type_name (Raw));
 			}
 		}
 
@@ -237,6 +294,8 @@ namespace GLib {
 		}
 
 		Hashtable before_signals;
+
+		[Obsolete ("Replaced by GLib.Signal marshaling mechanism.")]
 		protected Hashtable BeforeSignals {
 			get {
 				if (before_signals == null)
@@ -246,6 +305,7 @@ namespace GLib {
 		}
 
 		Hashtable after_signals;
+		[Obsolete ("Replaced by GLib.Signal marshaling mechanism.")]
 		protected Hashtable AfterSignals {
 			get {
 				if (after_signals == null)
@@ -255,6 +315,7 @@ namespace GLib {
 		}
 
 		EventHandlerList before_handlers;
+		[Obsolete ("Replaced by GLib.Signal marshaling mechanism.")]
 		protected EventHandlerList BeforeHandlers {
 			get {
 				if (before_handlers == null)
@@ -264,6 +325,7 @@ namespace GLib {
 		}
 
 		EventHandlerList after_handlers;
+		[Obsolete ("Replaced by GLib.Signal marshaling mechanism.")]
 		protected EventHandlerList AfterHandlers {
 			get {
 				if (after_handlers == null)
@@ -272,11 +334,60 @@ namespace GLib {
 			}
 		}
 
+		[CDeclCallback]
+		delegate void NotifyDelegate (IntPtr handle, IntPtr pspec, IntPtr gch);
+
+		void NotifyCallback (IntPtr handle, IntPtr pspec, IntPtr gch)
+		{
+			GLib.Signal sig = ((GCHandle) gch).Target as GLib.Signal;
+			if (sig == null)
+				throw new Exception("Unknown signal GC handle received " + gch);
+
+			NotifyArgs args = new NotifyArgs ();
+			args.Args = new object[1];
+			args.Args[0] = pspec;
+			NotifyHandler handler = (NotifyHandler) sig.Handler;
+			handler (GLib.Object.GetObject (handle), args);
+		}
+
+		void ConnectNotification (string signal, NotifyHandler handler)
+		{
+			Signal sig = Signal.Lookup (this, signal, new NotifyDelegate (NotifyCallback));
+			sig.AddDelegate (handler);
+		}
+
+		public void AddNotification (string property, NotifyHandler handler)
+		{
+			ConnectNotification ("notify::" + property, handler);
+		}
+
+		public void AddNotification (NotifyHandler handler)
+		{
+			ConnectNotification ("notify", handler);
+		}
+
+		void DisconnectNotification (string signal, NotifyHandler handler)
+		{
+			Signal sig = Signal.Lookup (this, signal, new NotifyDelegate (NotifyCallback));
+			sig.RemoveDelegate (handler);
+		}
+
+		public void RemoveNotification (string property, NotifyHandler handler)
+		{
+			DisconnectNotification ("notify::" + property, handler);
+		}
+
+		public void RemoveNotification (NotifyHandler handler)
+		{
+			DisconnectNotification ("notify", handler);
+		}
+
 		public override int GetHashCode ()
 		{
 			return Handle.GetHashCode ();
 		}
 
+		[Obsolete("Can cause instability due to garbage collection of GLib.Objects.")]
 		public Hashtable Data {
 			get { 
 				if (data == null)
@@ -286,36 +397,48 @@ namespace GLib {
 			}
 		}
 
+		protected Hashtable PersistentData {
+			get {
+				return WeakObject.Lookup (Handle).Data;
+			}
+		}
+
 		[DllImport("libgobject-2.0-0.dll")]
-		static extern void g_object_get_property (IntPtr obj, string name, ref GLib.Value val);
+		static extern void g_object_get_property (IntPtr obj, IntPtr name, ref GLib.Value val);
 
 		protected GLib.Value GetProperty (string name)
 		{
 			Value val = new Value (this, name);
-			g_object_get_property (Raw, name, ref val);
+			IntPtr native_name = GLib.Marshaller.StringToPtrGStrdup (name);
+			g_object_get_property (Raw, native_name, ref val);
+			GLib.Marshaller.Free (native_name);
 			return val;
 		}
 
 		[DllImport("libgobject-2.0-0.dll")]
-		static extern void g_object_set_property (IntPtr obj, string name, ref GLib.Value val);
+		static extern void g_object_set_property (IntPtr obj, IntPtr name, ref GLib.Value val);
 
 		protected void SetProperty (string name, GLib.Value val)
 		{
-			g_object_set_property (Raw, name, ref val);
+			IntPtr native_name = GLib.Marshaller.StringToPtrGStrdup (name);
+			g_object_set_property (Raw, native_name, ref val);
+			GLib.Marshaller.Free (native_name);
 		}
 
-		[DllImport("glibsharpglue")]
-		static extern void gtksharp_override_virtual_method (IntPtr gtype, string name, Delegate cb);
+		[DllImport("glibsharpglue-2")]
+		static extern void gtksharp_override_virtual_method (IntPtr gtype, IntPtr name, Delegate cb);
 
 		protected static void OverrideVirtualMethod (GType gtype, string name, Delegate cb)
 		{
-			gtksharp_override_virtual_method (gtype.Val, name, cb);
+			IntPtr native_name = GLib.Marshaller.StringToPtrGStrdup (name);
+			gtksharp_override_virtual_method (gtype.Val, native_name, cb);
+			GLib.Marshaller.Free (native_name);
 		}
 
 		[DllImport("libgobject-2.0-0.dll")]
 		protected static extern void g_signal_chain_from_overridden (IntPtr args, ref GLib.Value retval);
 
-		[DllImport("glibsharpglue")]
+		[DllImport("glibsharpglue-2")]
 		static extern bool gtksharp_is_object (IntPtr obj);
 
 		internal static bool IsObject (IntPtr obj)
@@ -323,7 +446,7 @@ namespace GLib {
 			return gtksharp_is_object (obj);
 		}
 
-		[DllImport("glibsharpglue")]
+		[DllImport("glibsharpglue-2")]
 		static extern int gtksharp_object_get_ref_count (IntPtr obj);
 
 		protected int RefCount {
